#pragma once

/*! Scalar value used. */
typedef float scalar;

#ifndef NULL
  /*! NULL value. */
  #define NULL 0
#endif

/*! The function was succesful. */
#define SUCCESS                        0
/*! Callistio is not yet initialized, use CAL_Initialisation. */
#define CAL_NOT_INIT                  -1
/*! The group does not exist. */
#define NO_SUCH_GROUP                 -2
/*! The parent group does not exist. */
#define NO_SUCH_PARENT_GROUP          -3
/*! The object does not exist. */
#define NO_SUCH_OBJECT                -4
/*! One of the parameters has an illegal value. */
#define ILLEGAL_VALUE                 -5
/*! Cannot collision check because both groups are part of the same tree-branch. */
#define GROUPS_IN_SAME_SUBTREE        -6
/*! This label does not exist. */
#define LABEL_NOT_FOUND               -7
/*! This label already exists. */
#define LABEL_ALREADY_EXISTS          -8
/*! The file you try to read is not valid or could not be found. */
#define FILE_ERROR                    -9
/*! You try to do something with the visualiser while it is not running. */
#define VIS_NOT_RUNNING              -10
/*! The clone you are trying to make cannot be put in the source's subtree. */
#define CLONE_IN_SUBGROUP            -11
/*! You try to do something with a view, while it is not visible. */
#define VIEW_NOT_VISIBLE             -12
/*! Cannot calculate penetration depth, because objects do not overlap. */
#define GROUPS_DO_NOT_OVERLAP        -13
/*! Group not capable of collision checks. */
#define GROUP_NOT_COL_CAPABLE        -14
/*! Cannot load texture, check filename/path and fileformat. */
#define TEXTURE_ERROR                -15

/*! This function initializes Callisto, starts the output window and GUI.
	\param visualisation Set to false if you dont want a visualisation window, default is TRUE.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_Initialisation(bool visualisation=1);

/*! This function ends Callisto and cleans up memory.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_End();

/*! This function stops the visualisation until CAL_ResumeVisualisation is called.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SuspendVisualisation();

/*! This function resumes the visualisation after CAL_SuspendVisualisation.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_ResumeVisualisation();

/*! Shows a view. Every view has its own unique ID.
    There are at most 4 views (0..3). 0 being the main view. This view cannot be switched on/off.
  \param viewID The ID of the view.
  \param title The title of the window of the view, default is no title.
  \returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_ShowView (int viewID, char* title="");

/*! Hides a view. Every view has its own unique ID.
    There are at most 4 views (0..3). 0 being the main view. This view cannot be switched on/off.
  \param viewID The ID of the view.
  \returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_HideView (int viewID);

/*! Shows or hides the visualisation axes.
  \param viewID The ID of the view to switch the axes on or off.
  \param visible Set to true to show the axes and to false to hide.
  \returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetAxesVisibility (int viewID, bool visible);

/*! This function sets the position and orientation of the camera in the visualisation window.
    It is positioned at eye and looks to the point defined with view. By default the up vector is (0,1,0)
    You can change this if necessary.
  \param viewID The ID of the view to set the view parameters for.
  \param eyeX The x position of the camera.
  \param eyeY The y position of the camera.
  \param eyeZ The z position of the camera.
  \param viewX The x position of the look at point.
  \param viewY The y position of the look at point.
  \param viewZ The z position of the look at point.
  \param upX The x direction of the up vector, default is 0.
  \param upY The y direction of the up vector, default is 1.
  \param upZ The z direction of the up vector, default is 0.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetViewParams (int viewID, scalar eyeX, scalar eyeY, scalar eyeZ, scalar viewX, scalar viewY, scalar viewZ, scalar upX=0, scalar upY=1, scalar upZ=0);

/*! This function gets the current position and orientation of the camera in a visualisation window.
    It is positioned at eye and looks to the point defined with view. 
  \param viewID The ID of the view get the view parameters from.
  \param *eyeX The x position of the camera.
  \param *eyeY The y position of the camera.
  \param *eyeZ The z position of the camera.
  \param *viewX The x position of the look at point.
  \param *viewY The y position of the look at point.
  \param *viewZ The z position of the look at point.
  \param *upX The x direction of the up vector, this is optional.
  \param *upY The y direction of the up vector, this is optional.
  \param *upZ The z direction of the up vector, this is optional.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_GetViewParams (int viewID, float *eyeX, float *eyeY, float *eyeZ, float *viewX, float *viewY, float *viewZ, float *upX=NULL, float *upY=NULL, float *upZ=NULL);

/*! Sets the background color of the visualisation window.
  \param viewID The ID of the view get the view parameters from.
  \param r The red component of the color (0...1).
  \param g The green component of the color (0...1).
  \param b The blue component of the color (0...1).
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetBackgroundColor (int viewID, float r, float g, float b);

/*! Creates an object group.
  \param groupID This is set to the group ID the group gets.
  \param parent The parent group of this group, use 0 if this group does not need a parent.
  \param colCheck True if this group is used for collision checks, false if not.
  \param label The label of the group as shown in the GUI, default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_DefineGroup (int* groupID, int parent, bool colCheck, char* label="");

/*! Delets a group, its child groups and all its objects.
  \param groupID The ID of the group to delete.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_DestroyGroup (int groupID);

/*! Deletes a groups child groups and all its objects.
  \param groupID The ID of the group to empty.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_EmptyGroup (int groupID);

/*! Destroys an object.
  \param objID The ID of the object to destroy.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_DestroyObject (int objID);

/*! Place a group and all its objects at a new position.
  \param groupID The ID of the group to translate.
  \param x The x component of the new position.
  \param y The y component of the new position.
  \param z The z component of the new position.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetGroupTranslation (int groupID, scalar x, scalar y, scalar z);

/*! Place a group and all its objects in a new orientation using Euler angles (in radians).
  \param groupID The ID of the group to translate.
  \param xRot The rotation about the x-axis;
  \param yRot The rotation about the y-axis;
  \param zRot The rotation about the z-axis;
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetGroupOrientation (int groupID, scalar xRot, scalar yRot, scalar zRot);

/*! Scale a group and all its objects.
  \param groupID The ID of the group to scale.
  \param xScale The scaling factor in the x direction.
  \param yScale The scaling factor in the y direction.
  \param zScale The scaling in the z direction.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetGroupScaling (int groupID, scalar xScale, scalar yScale, scalar zScale);

/*! Rotate a group and all its objects using a quaternion.
  \param groupID The ID of the group to rotate.
  \param x The x component of the quaternion.
  \param y The y component of the quaternion.
  \param z The z component of the quaternion.
  \param w The w component of the quaternion.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetGroupQuaternion (int groupID, scalar x, scalar y, scalar z, scalar w);

/*! Sets the color of an object.
  \param objID The ID of the object.
  \param r The red component of the color (0...1).
  \param g The green component of the color (0...1).
  \param b The blue component of the color (0...1).
  \param a The alpha value of the color (0...1. 0 is fully transparant, 1 is fully opaque.
  \param sID IGNORE
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetObjectColor (int objID, float r, float g, float b, float a=1, int sID=0);

/*! Sets the visibility of a group in a particular view. This does not effect collision checks.
  \param groupID The ID of the group.
  \param viewID The ID of the view.
  \param visible True if the group should be visible. False for invisibility.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_SetGroupVisibility (int groupID, int viewID, bool visible);

/*! Clones a group including all objects.
  \param groupIDNew Set to the group ID of the clone.
  \param groupID The ID of the group to clone.
  \param parent The parent group of the clone (0 for no parent).
  \param colCheck Set to false if clone group does not need collision checks.
  \param label The label of the group, default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CloneGroup (int* groupIDNew, int groupID, int parent, bool colCheck, char* label="");

/*! Returnes the ID of a group or object with a certain label.
  \param ID This is set to the ID.
  \param label The label of the group/object.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_GetID (int* ID, char* label);

/*! Create a box.
  \param groupID The group ID to put the object in.
  \param xw The width of the object.
  \param yw The height of the object.
  \param zw The depth of the object.
  \param x The x position of the object.
  \param y The y position of the object.
  \param z The z position of the object.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreateBox (int groupID, scalar xw, scalar yw, scalar zw, scalar x, scalar y, scalar z, int* objID=NULL, char* label="");

/*! Create a sphere.
  \param groupID The group ID to put the object in.
  \param r The radius of the object.
  \param x The x position of the object.
  \param y The y position of the object.
  \param z The z position of the object.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreateSphere (int groupID, scalar r, scalar x, scalar y, scalar z, int* objID=NULL, char* label="");

/*! Create a cylinder.
  \param groupID The group ID to put the object in.
  \param r The radius of the object.
  \param h The height of the object.
  \param x The x position of the object.
  \param y The y position of the object.
  \param z The z position of the object.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreateCylinder (int groupID, scalar r, scalar h, scalar x, scalar y, scalar z, int* objID=NULL, char* label="");

/*! Create a cone.
  \param groupID The group ID to put the object in.
  \param r The radius of the bottom of the object.
  \param h The height of the object.
  \param x The x position of the object.
  \param y The y position of the object.
  \param z The z position of the object.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreateCone (int groupID, scalar r, scalar h, scalar x, scalar y, scalar z, int* objID=NULL, char* label="");

/*! Create a triangle mesh.
  \param groupID The group ID to put the object in.
  \param nt The number of triangles the object consists of.
  \param *p List of coordinates of the object. The size of the list must be 3*nt.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreateTriangles (int groupID, int nt, scalar* p, int* objID=NULL, char* label="");

/*! Create a polygon. Must be planar and convex.
  \param groupID The group ID to put the object in.
  \param np The number of points the object consists of.
  \param *p List of coordinates of the object. The size of the list must be 3*np.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreatePolygon (int groupID, int np, scalar* p, int* objID=NULL, char* label="");

/*! Create a polyline. This is for the visualisation only.
  \param groupID The group ID to put the object in.
  \param nl The number of lines.
  \param *np The number of points each line consists of.
  \param *p List of coordinates of the lines. Its size should be is 3 * np * nl.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreatePolyline (int groupID, int nl, int *np, scalar *p, int* objID=NULL, char* label="");

/*! Create a tetrahedron.
  \param groupID The group ID to put the object in.
  \param *p List of coordinates of the object. The size of the list must be 3*4 since a tetrahydron consists of 4 points.
  \param objID Set to the object ID.
  \param label Label of the object. Default is no label.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CreateTetrahedron (int groupID, scalar* p, int* objID=NULL, char* label="");

/*! Check wheter a point collides with a group.
  \param groupID The group ID to check with.
  \param x The x coordinate of the point to check.
  \param y The y coordinate of the point to check.
  \param z The z coordinate of the point to check.
  \param multiple Flag whether to find all results, or just the first (faster).
  \param *nrCols The number of collisions.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CheckPointCollision (int groupID, scalar x, scalar y, scalar z, bool multiple, int *nrCols);

/*! Check wheter a line collides with a group.
  \param groupID The group ID to check with.
  \param x0 The x coordinate of the first point to check.
  \param y0 The y coordinate of the first point to check.
  \param z0 The z coordinate of the first point to check.
  \param x1 The x coordinate of the second point to check.
  \param y1 The y coordinate of the second point to check.
  \param z1 The z coordinate of the second point to check.
  \param multiple Flag whether to find all results, or just the first (faster).
  \param *nrCols The number of collisions.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CheckLineCollision (int groupID, scalar x0, scalar y0, scalar z0, scalar x1, scalar y1, scalar z1, bool multiple, int *nrCols);

/*! Check wheter two groups collide. Groups cannot be each others subgroups.
  \param group0 The ID of the first group.
  \param group1 The ID of the second group.
  \param multiple Flag whether to find all results, or just the first (faster).
  \param *nrCols The number of collisions.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_CheckGroupCollision (int group0, int group1, bool multiple, int *nrCols);

/*! Load a scene in XML format from disk.
  \param *fileName The name of the file to load.
  \param parent The parent group to put the loaded scene in, use 0 for no parent.
  \param *error String with possible error string. Can be omitted.
	\returns The function returns SUCCESS on success, and an errorcode on failure.
*/
int CAL_LoadXMLScene (char* fileName, int parent, char **error=NULL);

/*! \file callisto.h
*/